@@ -49,16 +49,61 @@ static int hnbgw_hnbap_tx(struct hnb_context *ctx, struct msgb *msg)
 	return 0;
 }
 
-static int hnbgw_tx_hnb_register_rej(struct hnb_context *ctx)
+static int hnbgw_tx_error_ind(struct hnb_context *ctx, const HNBAP_Cause_t *cause,
+			      HNBAP_ProcedureCode_t proc_code, HNBAP_Criticality_t criticality,
+			      HNBAP_TriggeringMessage_t trig_msg)
+{
+	HNBAP_ErrorIndication_t err_ind;
+	HNBAP_ErrorIndicationIEs_t err_ind_ies;
+	struct msgb *msg;
+	int rc;
+
+	LOGHNB(ctx, DHNBAP, LOGL_ERROR, "Tx ErrorIndication cause=%s\n", hnbap_cause_str(cause));
+
+	err_ind_ies.presenceMask = 0;
+	err_ind_ies.cause = *cause;
+
+	if (proc_code != -1 || trig_msg != -1 || criticality != -1) {
+		err_ind_ies.presenceMask |= ERRORINDICATIONIES_HNBAP_CRITICALITYDIAGNOSTICS_PRESENT;
+		if (proc_code != -1)
+			err_ind_ies.criticalityDiagnostics.procedureCode = &proc_code;
+		if (trig_msg != -1)
+			err_ind_ies.criticalityDiagnostics.triggeringMessage = &trig_msg;
+		if (criticality != -1)
+			err_ind_ies.criticalityDiagnostics.procedureCriticality = &criticality;
+	}
+
+	memset(&err_ind, 0, sizeof(err_ind));
+	rc = hnbap_encode_errorindicationies(&err_ind, &err_ind_ies);
+	if (rc < 0) {
+		LOGHNB(ctx, DHNBAP, LOGL_ERROR, "Failure to encode ERROR-INDICATION to %s: rc=%d\n",
+			ctx->identity_info, rc);
+		return rc;
+	}
+
+	msg = hnbap_generate_initiating_message(HNBAP_ProcedureCode_id_ErrorIndication,
+						HNBAP_Criticality_ignore,
+						&asn_DEF_HNBAP_ErrorIndication,
+						&err_ind);
+
+	ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_HNBAP_ErrorIndication, &err_ind);
+
+	return hnbgw_hnbap_tx(ctx, msg);
+}
+
+static int hnbgw_tx_hnb_register_rej(struct hnb_context *ctx, const HNBAP_Cause_t *cause)
 {
 	HNBAP_HNBRegisterReject_t reject_out;
 	HNBAP_HNBRegisterRejectIEs_t reject;
 	struct msgb *msg;
 	int rc;
 
-	reject.presenceMask = 0,
-	reject.cause.present = HNBAP_Cause_PR_radioNetwork;
-	reject.cause.choice.radioNetwork = HNBAP_CauseRadioNetwork_unspecified;
+	OSMO_ASSERT(cause);
+
+	LOGHNB(ctx, DHNBAP, LOGL_ERROR, "Rejecting HNB Register Request cause=%s\n", hnbap_cause_str(cause));
+
+	reject.presenceMask = 0;
+	reject.cause = *cause;
 
 	/* encode the Information Elements */
 	memset(&reject_out, 0, sizeof(reject_out));
@@ -161,13 +206,15 @@ static int hnbgw_tx_ue_register_acc(struct ue_context *ue)
 	return hnbgw_hnbap_tx(ue->hnb, msg);
 }
 
-static int hnbgw_tx_ue_register_rej_tmsi(struct hnb_context *hnb, HNBAP_UE_Identity_t *ue_id)
+static int hnbgw_tx_ue_register_rej(struct hnb_context *hnb, HNBAP_UE_Identity_t *ue_id, const struct HNBAP_Cause *cause)
 {
 	HNBAP_UERegisterReject_t reject_out;
 	HNBAP_UERegisterRejectIEs_t reject;
 	struct msgb *msg;
 	int rc;
 
+	OSMO_ASSERT(cause);
+
 	memset(&reject, 0, sizeof(reject));
 	reject.uE_Identity.present = ue_id->present;
 
@@ -229,10 +276,9 @@ static int hnbgw_tx_ue_register_rej_tmsi(struct hnb_context *hnb, HNBAP_UE_Ident
 		return -1;
 	}
 
-	LOGHNB(hnb, DHNBAP, LOGL_ERROR, "Rejecting UE Register Request: TMSI identity registration is switched off\n");
+	LOGHNB(hnb, DHNBAP, LOGL_ERROR, "Rejecting UE Register Request cause=%s\n", hnbap_cause_str(cause));
 
-	reject.cause.present = HNBAP_Cause_PR_radioNetwork;
-	reject.cause.choice.radioNetwork = HNBAP_CauseRadioNetwork_invalid_UE_identity;
+	reject.cause = *cause;
 
 	memset(&reject_out, 0, sizeof(reject_out));
 	rc = hnbap_encode_ueregisterrejecties(&reject_out, &reject);
@@ -388,11 +434,16 @@ static int hnbgw_tx_ue_register_acc_tmsi(struct hnb_context *hnb, HNBAP_UE_Ident
 static int hnbgw_rx_hnb_deregister(struct hnb_context *ctx, ANY_t *in)
 {
 	HNBAP_HNBDe_RegisterIEs_t ies;
+	HNBAP_Cause_t cause;
 	int rc;
 
 	rc = hnbap_decode_hnbde_registeries(&ies, in);
-	if (rc < 0)
-		return rc;
+	if (rc < 0) {
+		cause.present = HNBAP_Cause_PR_protocol;
+		cause.choice.radioNetwork = HNBAP_CauseProtocol_unspecified;
+		return hnbgw_tx_error_ind(ctx, &cause, HNBAP_ProcedureCode_id_HNBDe_Register, HNBAP_Criticality_ignore,
+					  HNBAP_TriggeringMessage_initiating_message);
+	}
 
 	LOGHNB(ctx, DHNBAP, LOGL_DEBUG, "HNB-DE-REGISTER cause=%s\n", hnbap_cause_str(&ies.cause));
 
@@ -413,11 +464,14 @@ static int hnbgw_rx_hnb_register_req(struct hnb_context *ctx, ANY_t *in)
 	char identity_str[256];
 	const char *cell_id_str;
 	struct timespec tp;
+	HNBAP_Cause_t cause;
 
 	rc = hnbap_decode_hnbregisterrequesties(&ies, in);
 	if (rc < 0) {
 		LOGHNB(ctx, DHNBAP, LOGL_ERROR, "Failure to decode HNB-REGISTER-REQ: rc=%d\n", rc);
-		return rc;
+		cause.present = HNBAP_Cause_PR_protocol;
+		cause.choice.radioNetwork = HNBAP_CauseProtocol_unspecified;
+		return hnbgw_tx_hnb_register_rej(ctx, &cause);
 	}
 	asn1_strncpy(identity_str, &ies.hnB_Identity.hNB_Identity_Info, sizeof(identity_str));
 
@@ -439,7 +493,9 @@ static int hnbgw_rx_hnb_register_req(struct hnb_context *ctx, ANY_t *in)
 	if (!hnbp) {
 		LOGHNB(ctx, DHNBAP, LOGL_NOTICE, "Rejecting unknonwn HNB with identity %s\n", identity_str);
 		hnbap_free_hnbregisterrequesties(&ies);
-		return hnbgw_tx_hnb_register_rej(ctx);
+		cause.present = HNBAP_Cause_PR_radioNetwork;
+		cause.choice.radioNetwork = HNBAP_CauseRadioNetwork_unauthorised_HNB;
+		return hnbgw_tx_hnb_register_rej(ctx, &cause);
 	}
 	ctx->persistent = hnbp;
 	hnbp->ctx = ctx;
@@ -479,7 +535,9 @@ static int hnbgw_rx_hnb_register_req(struct hnb_context *ctx, ANY_t *in)
 			 * misconfigurations or someone trying to impersonate an already working HNB: */
 			LOGHNB(ctx, DHNBAP, LOGL_ERROR, "rejecting HNB-REGISTER-REQ with duplicate cell identity %s\n", cell_id_str);
 			hnbap_free_hnbregisterrequesties(&ies);
-			return hnbgw_tx_hnb_register_rej(ctx);
+			cause.present = HNBAP_Cause_PR_radioNetwork;
+			cause.choice.radioNetwork = HNBAP_CauseRadioNetwork_hNB_parameter_mismatch;
+			return hnbgw_tx_hnb_register_rej(ctx, &cause);
 		}
 	}
 
@@ -503,14 +561,28 @@ static int hnbgw_rx_hnb_register_req(struct hnb_context *ctx, ANY_t *in)
 static int hnbgw_rx_ue_register_req(struct hnb_context *ctx, ANY_t *in)
 {
 	HNBAP_UERegisterRequestIEs_t ies;
+	HNBAP_Cause_t cause;
 	struct ue_context *ue;
 	struct ue_context *ue_allocated = NULL;
 	char imsi[16];
 	int rc;
 
 	rc = hnbap_decode_ueregisterrequesties(&ies, in);
-	if (rc < 0)
-		return rc;
+	if (rc < 0) {
+		LOGHNB(ctx, DHNBAP, LOGL_ERROR, "Failure to decode UE-REGISTER-REQ: rc=%d\n", rc);
+		cause.present = HNBAP_Cause_PR_protocol;
+		cause.choice.radioNetwork = HNBAP_CauseProtocol_unspecified;
+		return hnbgw_tx_error_ind(ctx, &cause, HNBAP_ProcedureCode_id_UERegister, HNBAP_Criticality_reject,
+					  HNBAP_TriggeringMessage_initiating_message);
+	}
+
+	if (!ctx->hnb_registered) {
+		/* UE registration requires prior HNB registration; reject with proper cause */
+		cause.present = HNBAP_Cause_PR_radioNetwork;
+		cause.choice.radioNetwork = HNBAP_CauseRadioNetwork_hNB_not_registered;
+		rc = hnbgw_tx_ue_register_rej(ctx, &ies.uE_Identity, &cause);
+		goto free_and_return_rc;
+	}
 
 	switch (ies.uE_Identity.present) {
 	case HNBAP_UE_Identity_PR_iMSI:
@@ -527,16 +599,21 @@ static int hnbgw_rx_ue_register_req(struct hnb_context *ctx, ANY_t *in)
 		break;
 	case HNBAP_UE_Identity_PR_tMSILAI:
 	case HNBAP_UE_Identity_PR_pTMSIRAI:
-		if (g_hnbgw->config.hnbap_allow_tmsi)
+		if (g_hnbgw->config.hnbap_allow_tmsi) {
 			rc = hnbgw_tx_ue_register_acc_tmsi(ctx, &ies.uE_Identity);
-		else
-			rc = hnbgw_tx_ue_register_rej_tmsi(ctx, &ies.uE_Identity);
+		} else {
+			cause.present = HNBAP_Cause_PR_radioNetwork;
+			cause.choice.radioNetwork = HNBAP_CauseRadioNetwork_invalid_UE_identity;
+			rc = hnbgw_tx_ue_register_rej(ctx, &ies.uE_Identity, &cause);
+		}
 		/* all has been handled by TMSI, skip the IMSI code below */
 		goto free_and_return_rc;
 	default:
 		LOGHNB(ctx, DHNBAP, LOGL_NOTICE, "UE-REGISTER-REQ with unsupported UE Id type %d\n",
 			ies.uE_Identity.present);
-		rc = -ENOTSUP;
+		cause.present = HNBAP_Cause_PR_radioNetwork;
+		cause.choice.radioNetwork = HNBAP_CauseRadioNetwork_invalid_UE_identity;
+		rc = hnbgw_tx_ue_register_rej(ctx, &ies.uE_Identity, &cause);
 		goto free_and_return_rc;
 	}
 
@@ -563,21 +640,34 @@ free_and_return_rc:
 static int hnbgw_rx_ue_deregister(struct hnb_context *ctx, ANY_t *in)
 {
 	HNBAP_UEDe_RegisterIEs_t ies;
+	HNBAP_Cause_t cause;
 	struct ue_context *ue;
 	int rc;
 	uint32_t ctxid;
 
 	rc = hnbap_decode_uede_registeries(&ies, in);
-	if (rc < 0)
-		return rc;
+	if (rc < 0) {
+		cause.present = HNBAP_Cause_PR_protocol;
+		cause.choice.protocol = HNBAP_CauseProtocol_unspecified;
+		return hnbgw_tx_error_ind(ctx, &cause, HNBAP_ProcedureCode_id_UEDe_Register,
+					  HNBAP_Criticality_ignore, HNBAP_TriggeringMessage_initiating_message);
+	}
 
 	ctxid = asn1bitstr_to_u24(&ies.context_ID);
 
-	LOGHNB(ctx, DHNBAP, LOGL_DEBUG, "UE-DE-REGISTER context=%u cause=%s\n", ctxid, hnbap_cause_str(&ies.cause));
-
-	ue = ue_context_by_id(ctxid);
-	if (ue)
-		ue_context_free(ue);
+	if (!ctx->hnb_registered) {
+		LOGHNB(ctx, DHNBAP, LOGL_NOTICE, "UE-DE-REGISTER context=%u cause=%s not permitted; HNB not registered\n",
+			ctxid, hnbap_cause_str(&ies.cause));
+		cause.present = HNBAP_Cause_PR_radioNetwork;
+		cause.choice.radioNetwork = HNBAP_CauseRadioNetwork_hNB_not_registered;
+		hnbgw_tx_error_ind(ctx, &cause, HNBAP_ProcedureCode_id_UEDe_Register,
+				   HNBAP_Criticality_ignore, HNBAP_TriggeringMessage_initiating_message);
+	} else {
+		LOGHNB(ctx, DHNBAP, LOGL_DEBUG, "UE-DE-REGISTER context=%u cause=%s\n", ctxid, hnbap_cause_str(&ies.cause));
+		ue = ue_context_by_id(ctxid);
+		if (ue)
+			ue_context_free(ue);
+	}
 
 	hnbap_free_uede_registeries(&ies);
 	return 0;
@@ -600,49 +690,40 @@ static int hnbgw_rx_err_ind(struct hnb_context *hnb, ANY_t *in)
 
 static int hnbgw_rx_initiating_msg(struct hnb_context *hnb, HNBAP_InitiatingMessage_t *imsg)
 {
+	HNBAP_Cause_t cause = {
+		.present = HNBAP_Cause_PR_protocol,
+		.choice.protocol = HNBAP_CauseProtocol_unspecified,
+	};
 	int rc = 0;
 
-	if (!hnb->hnb_registered) {
-		switch (imsg->procedureCode) {
-		case HNBAP_ProcedureCode_id_HNBRegister:	/* 8.2 */
-			rc = hnbgw_rx_hnb_register_req(hnb, &imsg->value);
-			break;
-		case HNBAP_ProcedureCode_id_HNBDe_Register:	/* 8.3 */
-			rc = hnbgw_rx_hnb_deregister(hnb, &imsg->value);
-			break;
-		default:
-			LOGHNB(hnb, DHNBAP, LOGL_NOTICE, "HNBAP Procedure %ld not permitted for de-registered HNB\n",
-				imsg->procedureCode);
-			break;
-		}
-	} else {
-		switch (imsg->procedureCode) {
-		case HNBAP_ProcedureCode_id_HNBRegister:	/*  8.2.4: Abnormal Condition, Accept. */
-			rc = hnbgw_rx_hnb_register_req(hnb, &imsg->value);
-			break;
-		case HNBAP_ProcedureCode_id_HNBDe_Register:	/* 8.3 */
-			rc = hnbgw_rx_hnb_deregister(hnb, &imsg->value);
-			break;
-		case HNBAP_ProcedureCode_id_UERegister:		/* 8.4 */
-			rc = hnbgw_rx_ue_register_req(hnb, &imsg->value);
-			break;
-		case HNBAP_ProcedureCode_id_UEDe_Register:	/* 8.5 */
-			rc = hnbgw_rx_ue_deregister(hnb, &imsg->value);
-			break;
-		case HNBAP_ProcedureCode_id_ErrorIndication:	/* 8.6 */
-			rc = hnbgw_rx_err_ind(hnb, &imsg->value);
-			break;
-		case HNBAP_ProcedureCode_id_TNLUpdate:		/* 8.9 */
-		case HNBAP_ProcedureCode_id_HNBConfigTransfer:	/* 8.10 */
-		case HNBAP_ProcedureCode_id_RelocationComplete:	/* 8.11 */
-		case HNBAP_ProcedureCode_id_U_RNTIQuery:	/* 8.12 */
-		case HNBAP_ProcedureCode_id_privateMessage:
-			LOGHNB(hnb, DHNBAP, LOGL_NOTICE, "Unimplemented HNBAP Procedure %ld\n", imsg->procedureCode);
-			break;
-		default:
-			LOGHNB(hnb, DHNBAP, LOGL_NOTICE, "Unknown HNBAP Procedure %ld\n", imsg->procedureCode);
-			break;
-		}
+	switch (imsg->procedureCode) {
+	case HNBAP_ProcedureCode_id_HNBRegister:	/* 8.2 */
+		rc = hnbgw_rx_hnb_register_req(hnb, &imsg->value);
+		break;
+	case HNBAP_ProcedureCode_id_HNBDe_Register:	/* 8.3 */
+		rc = hnbgw_rx_hnb_deregister(hnb, &imsg->value);
+		break;
+	case HNBAP_ProcedureCode_id_UERegister:		/* 8.4 */
+		rc = hnbgw_rx_ue_register_req(hnb, &imsg->value);
+		break;
+	case HNBAP_ProcedureCode_id_UEDe_Register:	/* 8.5 */
+		rc = hnbgw_rx_ue_deregister(hnb, &imsg->value);
+		break;
+	case HNBAP_ProcedureCode_id_ErrorIndication:	/* 8.6 */
+		rc = hnbgw_rx_err_ind(hnb, &imsg->value);
+		break;
+	case HNBAP_ProcedureCode_id_TNLUpdate:		/* 8.9 */
+	case HNBAP_ProcedureCode_id_HNBConfigTransfer:	/* 8.10 */
+	case HNBAP_ProcedureCode_id_RelocationComplete:	/* 8.11 */
+	case HNBAP_ProcedureCode_id_U_RNTIQuery:	/* 8.12 */
+	case HNBAP_ProcedureCode_id_privateMessage:
+		LOGHNB(hnb, DHNBAP, LOGL_NOTICE, "Unimplemented HNBAP Procedure %ld\n", imsg->procedureCode);
+		rc = hnbgw_tx_error_ind(hnb, &cause, imsg->procedureCode, -1, HNBAP_TriggeringMessage_initiating_message);
+		break;
+	default:
+		LOGHNB(hnb, DHNBAP, LOGL_NOTICE, "Unknown HNBAP Procedure %ld\n", imsg->procedureCode);
+		rc = hnbgw_tx_error_ind(hnb, &cause, imsg->procedureCode, -1, HNBAP_TriggeringMessage_initiating_message);
+		break;
 	}
 
 	return rc;
